#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import pathlib
from typing import Dict, List

# pylint: disable=C0114, C0115, C0116, R0903

from workspaces.capability.helpers_interfaces import ParameterIF
from workspaces.products.schema_interfaces import FutureProductIF
from workspaces.system.schema import JSONSerializable


class CapabilityIF(JSONSerializable):
    id: int
    name: str
    max_jobs: int

    def create_request(self, parameters: ParameterIF, future_products: FutureProductIF) -> "CapabilityRequestIF":
        raise NotImplementedError


class CapabilityRequestIF(JSONSerializable):
    id: int
    state: str
    capability: CapabilityIF
    parameters: str
    future_products: str
    versions: List[CapabilityVersionIF]

    @property
    def current_version(self):
        raise NotImplementedError


class CapabilityVersionIF:
    capability_request: CapabilityRequestIF
    version_number: int
    parameters: List[ParameterIF]
    workflow_metadata: str  # JSON-formatted
    files: List[CapabilityVersionFileIF]

    def __json__(self):
        raise NotImplementedError


class CapabilityVersionFileIF:
    capability_request_id: int
    version_number: int
    filename: str
    content: bytes

    def __json__(self):
        raise NotImplementedError


# TODO: This interface is slated for eventual removal
class CapabilityExecutionIF:
    id: int
    capability_version: CapabilityVersionIF
    state_name: str
    state: "State"
    version: CapabilityVersionIF
    current_workflow_request_id: int
    capability: CapabilityIF
    capability_request: CapabilityRequestIF
    delivery_url: str
    delivery_path: pathlib.Path

    def signal(self, ExecutionManagerIF, message: Dict):
        raise NotImplementedError

    def __json__(self):
        raise NotImplementedError


class ActionIF:
    """
    An action to take upon performing a transition. We expect to see several implementations
    of this interface:

    - SendNotification(template, additional_args) that sends a notification with
      the event and additional arguments

    - StartWorkflow(workflow_name, additional_args) that starts a workflow with the
      provided name, the event and additional arguments
    """

    def __call__(self, execution: CapabilityExecutionIF, manager):
        """
        Do something interesting during a state change.

        :param execution:
        :param manager:
        :return:
        """
        raise NotImplementedError

    def resume(self, execution: CapabilityExecutionIF, manager):
        """
        Resume work from this state, if we were queued for some reason
        :param execution: the execution in question
        :param manager:   the manager
        :return:
        """


class PatternIF:
    def matches(self, event: dict) -> bool:
        """
        This is most likely going to be implemented as a JSON Path expression, using
        the jsonpath-python library: https://pypi.org/project/jsonpath-python/

        :param event: a JSON object
        :return: true if this pattern matches that event object
        """
        raise NotImplementedError
